;; Generated package description from denote.el  -*- no-byte-compile: t -*-
(define-package "denote" "0.2.1" "Simple notes with an efficient file-naming scheme" '((emacs "27.2")) :commit "f10fe51fcd80bbcfc44bbef31d5b8f6952b4be4c" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Denote Development" . "~protesilaos/denote@lists.sr.ht") :url "https://git.sr.ht/~protesilaos/denote")
